"""
This module contains the user-facing API to submit PyTorch jobs to a Dask cluster
"""

import os
from typing import List, Callable, Tuple, Any
from dask.distributed import Client
import torch.distributed as dist


def _get_worker_info(client: Client) -> Tuple[List[str], str]:
    """
    returns a list of workers (sorted), and the DNS name for the master host
    The master is the 0th worker's host
    """
    workers = client.scheduler_info()["workers"]
    worker_keys = sorted(workers.keys())
    host = workers[worker_keys[0]]["host"]
    return worker_keys, host


def run(client: Client, pytorch_function: Callable, *args, **kwargs):
    """
    Dispatch a pytorch function over a dask cluster, and returns a list of futures
    for the resulting tasks
    """
    worker_keys, host = _get_worker_info(client)
    world_size = len(worker_keys)
    port = 23456  # pick a free port?

    futures = [
        client.submit(
            dispatch_with_ddp,
            pytorch_function=pytorch_function,
            master_addr=host,
            master_port=port,
            rank=idx,
            world_size=world_size,
            backend="nccl",
            *args,
            **kwargs
        )
        for idx, w in enumerate(worker_keys)
    ]

    return futures


# pylint: disable=keyword-arg-before-vararg
# pylint: disable=too-many-arguments
def dispatch_with_ddp(
    pytorch_function: Callable,
    master_addr: Any,
    master_port: Any,
    rank: Any,
    world_size: Any,
    backend: str = "nccl",
    *args,
    **kwargs
) -> Any:
    """
    runs a pytorch function, setting up torch.distributed before execution
    and tearing it down afterwards.
    """
    # These are the parameters used to initialize the process group
    master_addr = str(master_addr)
    master_port = str(master_port)
    rank = str(rank)
    world_size = str(world_size)

    os.environ["MASTER_ADDR"] = master_addr
    os.environ["MASTER_PORT"] = master_port
    os.environ["RANK"] = rank
    os.environ["WORLD_SIZE"] = world_size

    try:
        dist.init_process_group(backend=backend)
        val = pytorch_function(*args, **kwargs)
    finally:
        dist.destroy_process_group()
    return val



"""
Infrastructure for retrieving and logging intermediate results from pytorch training jobs.

Currently using dask pub/sub, but will create an S3 version in the future.
"""
import uuid
import logging
import os
from typing import List, Optional
from os.path import join, exists, dirname

from distributed.pubsub import Pub, Sub
from distributed.utils import TimeoutError as DistributedTimeoutError
from distributed.client import wait, FIRST_COMPLETED, Future


logger = logging.getLogger(__name__)


class DaskResultsHandler:
    """
    This class use Dask pubsub infra to pass intermediate results back from PyTorch
    jobs to the client.
    """

    def __init__(self, pub_sub_key: Optional[str] = None):
        """
        pub_sub_key is an arbitrary string (topic) for the pub sub channel.
        It's a good idea to change it.  Sometimes old topics can get "clogged"
        """
        if pub_sub_key is None:
            pub_sub_key = uuid.uuid4().hex
        self.pub_sub_key = pub_sub_key

    @classmethod
    def _get_all(cls, sub: Sub):
        while True:
            try:
                yield sub.get(timeout=1.0)
            except DistributedTimeoutError:
                break

    def _get_results(self, futures: List[Future], raise_errors: bool = True):
        sub = Sub(self.pub_sub_key)
        while True:
            for obj in self._get_all(sub):
                yield obj
            if not futures:
                break
            try:
                result = wait(futures, 0.1, FIRST_COMPLETED)
            except DistributedTimeoutError:
                continue

            for fut in result.done:
                try:
                    fut.result()
                except Exception as e:  # pylint: disable=broad-except
                    logging.exception(e)
                    if raise_errors:
                        raise
            futures = result.not_done

    def process_results(
        self, prefix: str, futures: List[Future], raise_errors: bool = True
    ) -> None:
        """
        Process the intermediate results:
        result objects will be dictionaries of the form {'path': path, 'data': data}
        As results come in, data will be written to f"prefix/{path}"

        prefix:  directory where you want results to be written
        futures:  list of futures for your jobs (output of dask_pytorch_ddp.dispatch.run)
        raise_errors:  If any of the jobs fail, either raise an exception, or log it and continue.
        """
        for result in self._get_results(futures, raise_errors=raise_errors):
            path = result["path"]
            data = result["data"]
            fpath = join(prefix, path)
            if not exists(dirname(fpath)):
                os.makedirs(dirname(fpath))
            if isinstance(data, str):
                data = data.encode("utf-8")
            with open(fpath, "wb+") as f:
                f.write(data)

    def submit_result(self, path: str, data: str):
        """
        To be used in jobs.  Call this function with a path, and some data.
        Client will write {data} to a file at {path}
        """
        pub = Pub(self.pub_sub_key)
        pub.put({"path": path, "data": data})



"""
Utilities for loading PyTorch data in distributed environments
"""


import tempfile
from os.path import basename, dirname
from typing import List, Callable, Optional
from PIL import Image
from torch.utils.data import Dataset


"""
In the following, we are explicitly avoiding s3fs because it does not behave well with
multiprocessing (which is commonly used in PyTorch dataloaders).

https://github.com/dask/s3fs/issues/369
"""  # pylint: disable=pointless-string-statement


def _list_all_files(bucket: str, prefix: str, s3_client=None, anon=False) -> List[str]:
    """
    Get list of all files from an s3 bucket matching a certain prefix
    """
    import boto3  # pylint: disable=import-outside-toplevel
    from botocore import UNSIGNED  # pylint: disable=import-outside-toplevel
    from botocore.client import Config  # pylint: disable=import-outside-toplevel

    if s3_client is None:
        if anon:
            s3_client = boto3.client("s3", config=Config(signature_version=UNSIGNED))
        else:
            s3_client = boto3.client("s3")

    paginator = s3_client.get_paginator("list_objects")
    all_files = []
    for page in paginator.paginate(Bucket=bucket, Prefix=prefix):
        files = [x["Key"] for x in page["Contents"]]
        all_files.extend(files)
    return all_files


def _read_s3_fileobj(bucket, path, fileobj, anon=False):
    """
    read an obj from s3 to a file like object
    """
    import boto3  # pylint: disable=import-outside-toplevel
    from botocore import UNSIGNED  # pylint: disable=import-outside-toplevel
    from botocore.client import Config  # pylint: disable=import-outside-toplevel

    if anon:
        s3 = boto3.resource("s3", config=Config(signature_version=UNSIGNED))
    else:
        s3 = boto3.resource("s3")

    bucket = s3.Bucket(bucket)
    bucket.download_fileobj(path, fileobj)
    fileobj.seek(0)
    return fileobj


def _load_image_obj(fileobj):
    """
    turn a file like object into an image
    """
    return Image.open(fileobj).convert("RGB")


class S3ImageFolder(Dataset):
    """
    An image folder that lives in S3.  Directories containing the image are classes.
    """

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-arguments

    def __init__(
        self,
        s3_bucket: str,
        s3_prefix: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        anon: Optional[bool] = False,
    ):
        self.s3_bucket = s3_bucket
        self.s3_prefix = s3_prefix
        self.anon = anon
        self.all_files = _list_all_files(s3_bucket, s3_prefix, anon=anon)
        self.classes = sorted({self._get_class(x) for x in self.all_files})
        self.class_to_idx = {k: idx for idx, k in enumerate(self.classes)}
        self.transform = transform
        self.target_transform = target_transform

    @classmethod
    def _get_class(cls, path):
        """
        parse the path to extract the class name
        """
        return basename(dirname(path))

    def __getitem__(self, idx):
        """
        get the nth (idx) image and label
        """
        path = self.all_files[idx]
        label = self.class_to_idx[self._get_class(path)]
        with tempfile.TemporaryFile() as f:
            f = _read_s3_fileobj(self.s3_bucket, path, f, self.anon)
            img = _load_image_obj(f)
        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            label = self.target_transform(label)
        return img, label

    def __len__(self):
        """
        total number of images
        """
        return len(self.all_files)
